/*
 *  linux/arch/arm/kernel/head-armv-s3c44b0x.S
 *
 *  Copyright (C) 1994-1999 Russell King
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 *  32-bit kernel startup code for S3C44B0X rom XIP config
 *  We don't use a bootldr for rom config, therefore we need to init hardware here.
 *
 *  by Martin Siu
 *
 */
#include <linux/config.h>
#include <linux/linkage.h>

#include <asm/assembler.h>
#include <asm/mach-types.h>
#include <asm/mach/arch.h>

#define BASE_1C     0x01c00000
#define SYSCFG	    0x00
#define NCACHBE0    0x04
#define NCACHBE1    0x08
#define MEMCON_BASE 0x00080000
#define BWSCON      0x00
#define BANKCON0    0x04
#define BANKCON1    0x08
#define BANKCON2    0x0c
#define BANKCON3    0x10
#define BANKCON4    0x14
#define BANKCON5    0x18
#define BANKCON6    0x1c
#define BANKCON7    0x20
#define REFRESH     0x24
#define BANKSIZE    0x28
#define MRSRB6      0x2c
#define MRSRB7      0x30

#define BASE_1D     0x01d00000
#define WTCON	    0x00030000  
#define PORT_BASE   0x00020000
#define rPCONA      0x00
#define rPDATA 	   	0x04
#define rPCONB     	0x08
#define rPDATB     	0x0c
#define rPCONC     	0x10
#define rPDATC     	0x14
#define rPUPC      	0x18
#define rPCOND     	0x1c
#define rPDATD     	0x20
#define rPUPD      	0x24
#define rPCONE     	0x28
#define rPDATE     	0x2c
#define rPUPE      	0x30
#define rPCONF     	0x34
#define rPDATF     	0x38
#define rPUPF      	0x3c
#define rPCONG     	0x40
#define rPDATG     	0x44
#define rPUPG      	0x48
#define rSPUCR     	0x4c
#define rEXTINT    	0x50

#define CLOCK_BASE  0x00080000
#define PLLCON	    0x00
#define CLKCON	    0x04
#define LOCKTIME    0x0c

#define BASE_1E     0x01e00000
#define INTPND	    0x04
#define INTMOD	    0x08
#define INTMSK	    0x0c
#define I_ISPR	    0x20
#define I_CMST	    0x1c

#define K(a,b,c)	((a) << 24 | (b) << 12 | (c))

/*
 *  Kernel startup entry point.
 *
 * The rules are:
 *  r0      - should be 0
 *  r1      - unique architecture number
 *  MMU     - off
 *  I-cache - on or off
 *  D-cache - off
 *
 * See linux/arch/arm/tools/mach-types for the complete list of numbers
 * for r1.
 */

.section ".text.init",#alloc,#execinstr
		.type	stext, #function
ENTRY(stext)

		b   hw_init			    /* 0x00 reset vector */
		b	vector_undefinstr	/* 0x04 undefined instruction  */
		b	vector_swi		    /* 0x08 software interrupt */
		b	vector_prefetch		/* 0x0C prefetch abort */
		b	vector_data		    /* 0x10 data abort */
		b	vector_addrexcptn	/* 0x14 address exception */
		b	vector_IRQ		    /* 0x18 IRQ */
		b	vector_FIQ		    /* 0x1C FIQ */
hw_init:

        mov	r12, r0

		mov	r0, #F_BIT | I_BIT | MODE_SVC	@ make sure svc mode
		msr	cpsr_c, r0			@ and all irqs disabled

        mov     r0, #BASE_1C
        mov     r1, #BASE_1D
        mov     r2, #BASE_1E
        add     r3, r1, #PORT_BASE
        add     r4, r1, #CLOCK_BASE
        add     r5, r0, #MEMCON_BASE

        // watch dog disable  
        // WTCON = 0x0 		   
        add     r6, r1, #WTCON
        mov     r7, #0x0
        str     r7, [r6]

        // SYSCFG = 0xe, enable writer buffer, full cache enable, stall disable 	
        add     r6, r0, #SYSCFG	    
        mov     r7, #0xe           	
        str     r7, [r6]

        // NCACHBE0	= 0x80002000   
        // bank1,2,3 area           		
        add     r6, r0, #NCACHBE0	    
        mov     r7, #0x80000000
        add     r7, r7, #0x2000   
        str     r7, [r6]

        // rPCONA = 0x3ff
        add	    r6, r3, #rPCONA
	    mov	    r7, #0x3f0
        add     r7, r7, #0xf
	    str	    r7, [r6]

        // rPCONB = 0x1ff
        add	    r6, r3, #rPCONB
	    mov	    r7, #0x1f0
        add     r7, r7, #0xf
	    str	    r7, [r6]

        // rPDATB = 0x0
        add	    r6, r3, #rPDATB
	    mov	    r7, #0x0
	    str	    r7, [r6]

        // rPCONC = 0xff55ff15
        add	    r6, r3, #rPCONC
	    mov	    r7, #0xff000000
        add     r7, r7, #0x00550000
        add     r7, r7, #0x0000ff00
        add     r7, r7, #0x00000015
	    str	    r7, [r6]

	    // rPDATC = 0x0    
        add	    r6, r3, #rPDATC
	    mov	    r7, #0x0
	    str	    r7, [r6]

	    // rPUPC = 0xffff               
        add	    r6, r3, #rPUPC
	    mov	    r7, #0xff00
        add     r7, r7, #0xff
	    str	    r7, [r6]

        // rPCOND = 0xaaaa
        add	    r6, r3, #rPCOND
	    mov	    r7, #0xaa00
        add     r7, r7, #0xaa
	    str	    r7, [r6]

        // rPUPD = 0xff
        add	    r6, r3, #rPUPD
	    mov	    r7, #0xff
	    str	    r7, [r6]

        // rPCONE = 0x200ab
        add	    r6, r3, #rPCONE
	    mov	    r7, #0x20000
        add     r7, r7, #0xab
	    str	    r7, [r6]

	    // rPDATE = 0x0
        add	    r6, r3, #rPDATE
	    mov	    r7, #0x0
	    str	    r7, [r6]

	    // rPUPE = 0xff 
        add	    r6, r3, #rPUPE
	    mov	    r7, #0xff
	    str	    r7, [r6]

        // rPCONF = 0x827ea
        add	    r6, r3, #rPCONF
	    mov	    r7, #0x80000
        add     r7, r7, #0x2700
        add     r7, r7, #0xea
	    str	    r7, [r6]

   	    // rPDATF = 0x63
        add	    r6, r3, #rPDATF
	    mov	    r7, #0x63
	    str	    r7, [r6]

	    // rPUPF = 0x1e3
        add	    r6, r3, #rPUPF
	    mov	    r7, #0x100
        add     r7, r7, #0xe3
	    str	    r7, [r6]

        // HIGH NIBBLE IIS MODE ,LOW NIBBLE EXINT INPUT MODE
	    // rPCONG = 0xaaff  
        add	    r6, r3, #rPCONG
	    mov	    r7, #0xaa00
        add     r7, r7, #0xff
	    str	    r7, [r6]

        // PORTG  EXTINT4 PULL-UP DISPABLE
	    // rPUPG = 0x0  
        add	    r6, r3, #rPUPG
	    mov	    r7, #0x0
	    str	    r7, [r6]   

        // SET ETC, DATA BUS pull-up disable    
	    // rSPUCR = 0x7
        add	    r6, r3, #rSPUCR
	    mov	    r7, #0x7
	    str	    r7, [r6]   

        // EINT1,0 falling edge, EINT2 rising edge triggered, EINT3 low level.
	    // rEXTINT = 0x422
        add	    r6, r3, #rEXTINT
	    mov	    r7, #0x400
        add	    r7, r7, #0x22
	    str	    r7, [r6]   


        // all interrupt disable
        // INTMSK = 0x07ffffff  
        add	    r6, r2, #INTMSK
	    mov	    r7, #0x07000000
        add	    r7, r7, #0x00ff0000
        add	    r7, r7, #0x0000ff00
        add	    r7, r7, #0x000000ff
	    str	    r7, [r6]   

        // Set clock control registers			            
        // EX) count = t_lock * Fin (t_lock=200us, Fin=10MHz) = 2000
        // LOCKTIME = 800	  
        add	    r6, r4, #LOCKTIME
        mov     r7, #800
        str     r7, [r6]

        // temporary setting of PLL
        // Fin=10MHz,Fout=60MHz
        // PLLCON = 0x34031     
        
        // Fin=10MHz,Fout=66MHz
        // PLLCON = 0x3A031     
        add	    r6, r4, #PLLCON
        //mov     r7, #0x34000
        mov     r7, #0x3A000
        add     r7, r7, #0x31
        str     r7, [r6]

        // All unit block CLK enable	
        // CLKCON = 0x7ff8	    
        add	    r6, r4, #CLKCON
        mov     r7, #0x7f00
        add     r7, r7, #0xf8
        str     r7, [r6]


        // Set memory control registers			        

        // BWSCON = 0x11000000
        add	    r6, r5, #BWSCON
        mov     r7, #0x11000000
        str     r7, [r6]

        // BANKCON0 = 0x600       
        add	    r6, r5, #BANKCON0
        mov     r7, #0x600
        str     r7, [r6]

        // BANKCON1 = 0x7ffc      
        add	    r6, r5, #BANKCON1
        mov     r7, #0x7f00
        add     r7, r7, #0xfc
        str     r7, [r6]

        // BANKCON2 = 0x7ffc      
        add	    r6, r5, #BANKCON2
        mov     r7, #0x7f00
        add     r7, r7, #0xfc
        str     r7, [r6]

        // BANKCON3 = 0x7ffc
        add	    r6, r5, #BANKCON3
        mov     r7, #0x7f00
        add     r7, r7, #0xfc
        str     r7, [r6]

        // BANKCON4 = 0x7ffc
        add	    r6, r5, #BANKCON4
        mov     r7, #0x7f00
        add     r7, r7, #0xfc
        str     r7, [r6]

        // BANKCON5 = 0x7ffc
        add	    r6, r5, #BANKCON5
        mov     r7, #0x7f00
        add     r7, r7, #0xfc
        str     r7, [r6]

        // BANKCON6 = 0x18001
        add	    r6, r5, #BANKCON6
        mov     r7, #0x18000
        add     r7, r7, #0x1  // disable for 2MB ram
        str     r7, [r6]

        // BANKCON7 = 0x18000
        add	    r6, r5, #BANKCON7
        mov     r7, #0x18000
        str     r7, [r6]

        // REFRESH = 0x82062b    
        add	    r6, r5, #REFRESH
        mov     r7, #0x820000
        add     r7, r7, #0x0600
        add     r7, r7, #0x2b
        str     r7, [r6]

        // BANKSIZE = 0x10   
        add	    r6, r5, #BANKSIZE
        mov     r7, #0x10
        str     r7, [r6]

        // MRSRB6 = 0x20 
        add	    r6, r5, #MRSRB6
        mov     r7, #0x20
        str     r7, [r6]

        // MRSRB7 = 0x20
        add	    r6, r5, #MRSRB7
        mov     r7, #0x20
        str     r7, [r6]

        adr     r2, LC0                                

        @  Copy data sections to their new home        
        ldmia   r2, {r4-r11,sp}                        
        cmp     r9, r11                                
        beq     2f                                     
1:      cmp     r11, r10                               
        ldrcc   r3, [r9], #4                           
        strcc   r3, [r11], #4                          
        bcc     1b                                     
                                
		@ clear BSS
2:              mov     r3, #0
3:              cmp     r7, r8
                strcc   r3, [r7], #4
                bcc     3b

		ldr	r3, S3C44B0X_PROCESSOR_TYPE
		str	r3, [r5]
		
		mov	r3, #MACH_TYPE_S3C44B0X
		str	r3, [r6]

        mov     fp, #0
        b       start_kernel


S3C44B0X_PROCESSOR_TYPE:
		.long	0x44B044B0
		
LC0:
        .long   0x07ff0000              @ r4
        .long   processor_id            @ r5
        .long   __machine_arch_type     @ r6
        .long   __bss_start             @ r7   bss start
        .long   _end                    @ r8   bss end


        .long   _etext_align            @ r9   data start in ROM (src)
        .long   _edata                  @ r10  data end in RAM
        .long   __data_start            @ r11  data start in RAM (dst)
        .long   init_task_union+8192    @ sp


		bl	__lookup_processor_type
		teq	r10, #0				@ invalid processor?
		moveq	r0, #'p'			@ yes, error 'p'
		beq	__error
		bl	__lookup_architecture_type
#if !defined(CONFIG_ARCTURUS)
		teq	r7, #0				@ invalid architecture?
		moveq	r0, #'a'			@ yes, error 'a'
		beq	__error
#endif
		adr	lr, __ret			@ return address
		add	pc, r10, #12			@ initialise processor
							@ (return control reg)

__switch_data:	.long	__mmap_switched
		.long	SYMBOL_NAME(compat)
		.long	SYMBOL_NAME(__bss_start)
		.long	SYMBOL_NAME(_end)
		.long	SYMBOL_NAME(processor_id)
		.long	SYMBOL_NAME(__machine_arch_type)
		.long	SYMBOL_NAME(cr_alignment)
		.long	SYMBOL_NAME(init_task_union)+8192

__ret:		ldr	lr, __switch_data
#ifdef CONFIG_CPU_WITH_CACHE
# ifdef CONFIG_CPU_WITH_MCR_INSTRUCTION
		mcr	p15, 0, r0, c1, c0
		mov	r0, r0
		mov	r0, r0
		mov	r0, r0
# else
#  warning 	"FIXME: Enable Cache, Other settings without MCR Instruction"
		@ Note r0 is initialized suitably to enable Cache
		@ and additional settings if any in proc_armX.S file
		@ A possible code here
		@ldr	r2, CACHE_CONTROL_MEM_LOCATION
		@str	r0, [r2]
# endif
#endif
		mov	pc, lr

		/*
		 * This code follows on after the page
		 * table switch and jump above.
		 *
		 * r0  = processor control register
		 * r1  = machine ID
		 * r9  = processor ID
		 */
		.align	5
__mmap_switched:
		adr	r3, __switch_data + 4
		ldmia	r3, {r2, r4, r5, r6, r7, r8, sp}@ r2 = compat
							@ sp = stack pointer
		str	r12, [r2]

		mov	fp, #0				@ Clear BSS (and zero fp)
1:		cmp	r4, r5
		strcc	fp, [r4],#4
		bcc	1b

		str	r9, [r6]			@ Save processor ID
		str	r1, [r7]			@ Save machine type
#ifdef CONFIG_ALIGNMENT_TRAP
		orr	r0, r0, #2			@ ...........A.
#endif
		bic	r2, r0, #2			@ Clear 'A' bit
		stmia	r8, {r0, r2}			@ Save control register values
		b	SYMBOL_NAME(start_kernel)




/*
 * Exception handling.  Something went wrong and we can't
 * proceed.  We ought to tell the user, but since we
 * don't have any guarantee that we're even running on
 * the right architecture, we do virtually nothing.
 * r0 = ascii error character:
 *	a = invalid architecture
 *	p = invalid processor
 *	i = invalid calling convention
 *
 * Generally, only serious errors cause this.
 */
__error:


/*
 * Read processor ID register (CP#15, CR0), and look up in the linker-built
 * supported processor list.  Note that we can't use the absolute addresses
 * for the __proc_info lists since we aren't running with the MMU on
 * (and therefore, we are not in the correct address space).  We have to
 * calculate the offset.
 *
 * Returns:
 *	r5, r6, r7 corrupted
 *	r8  = page table flags
 *	r9  = processor ID
 *	r10 = pointer to processor structure
 */
__lookup_processor_type:
		adr	r5, 2f
		ldmia	r5, {r7, r9, r10}
		sub	r5, r5, r10			@ convert addresses
		add	r7, r7, r5			@ to our address space
		add	r10, r9, r5
#ifdef CONFIG_CPU_WITH_MCR_INSTRUCTION
		mrc	p15, 0, r9, c0, c0		@ get processor id
#else
# warning 	"FIXME: Get Processor ID without MCR Instruction"
		@ A possible code
		@ldr	r9, PROCESSOR_ID_MEM_LOCATION
		@ldr	r9, [r9]
#endif
1:		ldmia	r10, {r5, r6, r8}		@ value, mask, mmuflags
		and	r6, r6, r9			@ mask wanted bits
		teq	r5, r6
		moveq	pc, lr
		add	r10, r10, #36			@ sizeof(proc_info_list)
		cmp	r10, r7
		blt	1b
		mov	r10, #0				@ unknown processor
		mov	pc, lr

/*
 * Look in include/asm-arm/procinfo.h and arch/arm/kernel/arch.[ch] for
 * more information about the __proc_info and __arch_info structures.
 */
2:		.long	__proc_info_end
		.long	__proc_info_begin
		.long	2b
		.long	__arch_info_begin
		.long	__arch_info_end

/*
 * Lookup machine architecture in the linker-build list of architectures.
 * Note that we can't use the absolute addresses for the __arch_info
 * lists since we aren't running with the MMU on (and therefore, we are
 * not in the correct address space).  We have to calculate the offset.
 *
 *  r1 = machine architecture number
 * Returns:
 *  r2, r3, r4 corrupted
 *  r5 = physical start address of RAM
 *  r6 = physical address of IO
 *  r7 = byte offset into page tables for IO
 */
__lookup_architecture_type:
		adr	r4, 2b
		ldmia	r4, {r2, r3, r5, r6, r7}	@ throw away r2, r3
		sub	r5, r4, r5			@ convert addresses
		add	r4, r6, r5			@ to our address space
		add	r7, r7, r5
1:		ldr	r5, [r4]			@ get machine type
		teq	r5, r1
		beq	2f
		add	r4, r4, #SIZEOF_MACHINE_DESC
		cmp	r4, r7
		blt	1b
		mov	r7, #0				@ unknown architecture
		mov	pc, lr
2:		ldmib	r4, {r5, r6, r7}		@ found, get results
		mov	r7, r7, lsr #18			@ pagetable byte offset
		mov	pc, lr




