#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/poll.h> 
#include <linux/s3c44b0x_pm.h> 
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/arch/io.h>
#include <asm/arch/irqs.h>
#include <asm/arch/s3c44b0x_iodirect.h>

#define __author__ "Martin Siu"
#define __description__ "Power Management for Toymaker video player"
#define __license__ "GPL"
#define __driver_name__ "s3c44b0x_pm"

static int s3c44b0x_pm_ioctl (struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
    switch (cmd)
    {
        case S3C44B0X_PM_IOC_POWER_OFF:
            // stop all clocks
            IO_S3C44B0X_CLKCON &= ~((1 << 14) | // I2S
                                    (1 << 13) | // I2C
                                    (1 << 12) | // ADC
                                    (1 << 11) | // RTC
                                    (1 << 10) | // GPIO
                                    (1 << 9)  | // Uart1
                                    (1 << 8)  | // Uart0
                                    (1 << 7)  | // BDMA
                                    (1 << 6)  | // LCDC
                                    (1 << 5)  | // SIO
                                    (1 << 4)  | // ZDMA
                                    (1 << 3));  // PWM
            udelay(100);                    
            // transition to STOP mode
            // CPU hang up from this onwards                
            IO_S3C44B0X_CLKCON |= (1 << 0); 
            return 0;

        default:
            return -EINVAL;            
    }
}

static struct file_operations s3c44b0x_pm_fops = 
{
    ioctl:      s3c44b0x_pm_ioctl,
};

int __init s3c44b0x_pm_init (void)
{    
    int err;
    
    if ((err = register_chrdev(104, __driver_name__, &s3c44b0x_pm_fops)))
		panic("Unable to register driver %s, err %d\n", __driver_name__, err);
        
	printk(KERN_INFO "%s\n", __description__);

	return 0;
}

__initcall(s3c44b0x_pm_init);
